import io
import os
from typing import Union

from ms_deisotope.data_source.infer_type import FormatGuesser, is_random_access
from ms_deisotope.data_source import _compression
from ms_deisotope.data_source.mzmlb import is_mzmlb_file as _is_mzmlb_file

from ms_deisotope.output.mzml import ProcessedMzMLLoader, MzMLSerializer
from ms_deisotope.output.mzmlb import ProcessedMzMLbLoader, MzMLbSerializer
from ms_deisotope.output.mgf import ProcessedMGFLoader, MGFSerializer

ProcessedMSFileLoader = FormatGuesser([], [ProcessedMzMLLoader, ProcessedMzMLbLoader, ProcessedMGFLoader])

ProcessedMSFileLoader.add_file_extension_guesser({
    ".mzml": ProcessedMzMLLoader,
    ".mzmlb": ProcessedMzMLbLoader,
    ".mgf": ProcessedMGFLoader
})


@ProcessedMSFileLoader.register_type_guesser
def guess_type_from_file_sniffing(file_path: Union[os.PathLike, io.IOBase]):
    is_random_access_file = is_random_access(file_path)
    if is_random_access_file:
        handle = file_path
        header = handle.read(1000)
        handle.seek(0)
    else:
        with open(file_path, 'rb') as handle:
            header = handle.read(1000)

    if _compression.starts_with_gz_magic(header):
        if is_random_access_file:
            handle = _compression.GzipFile(fileobj=file_path, mode='rb')
            header = handle.read(1000)
            handle.seek(0)
        else:
            with _compression.GzipFile(file_path, mode='rb') as handle:
                header = handle.read(1000)

    if b"mzML" in header:
        return ProcessedMzMLLoader
    elif b"BEGIN IONS" in header:
        return ProcessedMGFLoader
    else:
        raise ValueError("Cannot determine ScanLoader type from header")


@ProcessedMSFileLoader.register_type_guesser
def is_mzmlb_file(file_path):
    if _is_mzmlb_file(file_path):
        return ProcessedMzMLbLoader
    raise ValueError("Not an mzMLb file")


serializers = {}

for tp in [MzMLSerializer, MzMLbSerializer, MGFSerializer]:
    for ext in tp.file_extensions:
        serializers[ext] = tp


def get_writer(filename, **kwargs):
    """Open a writer for a provided filename, inferring the format from
    the file extension.

    .. warning::
        If using a file-like object, do not use a compressed writer or else
        the stream will be doubly-compressed.

    Parameters
    ----------
    filename : :class:`str`, :class:`os.PathLike`, or file-like object
        The path to the file to open, or a file-like object with a "name"
        attribute.
    **kwargs
        Keyword arguments forwarded to the writer

    Returns
    -------
    MzMLSerializer or MGFSerializer or MzMLbSerializer
    """
    if hasattr(filename, 'name'):
        handle = filename
        name = handle.name
    else:
        handle = None
        name = os.path.basename(filename)

    is_gzipped = False
    if name.lower().endswith(".gz"):
        is_gzipped = True
        name = name[:-3]

    name, ext = os.path.splitext(name.lower())
    serializer_cls = serializers[ext[1:]]

    if serializer_cls == MzMLbSerializer:
        return serializer_cls(filename, **kwargs)

    if is_gzipped:
        if handle is None:
            handle = _compression.GzipFile(filename, 'wb')
        else:
            handle = _compression.GzipFile(fileobj=filename, mode='wb')

    else:
        if handle is None:
            handle = open(filename, 'wb')

    return serializer_cls(handle, **kwargs)



from .utils import GeneratorQueue
from .intervals import SpanningMixin


def ident(x):
    return x


class FitNode(SpanningMixin):
    def __init__(self, fit, index=None):
        self.index = index
        self.fit = fit
        self.edges = set()
        self.overlap_edges = set()
        # Placeholder Peaks have a peak_count of -1
        self.peak_indices = {p.peak_count for p in fit.experimental if p.peak_count >= 0}
        self._hash = None
        self.score = fit.score
        self.start = fit.experimental[0].mz
        self.end = fit.experimental[-1].mz

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self.fit)
        return self._hash

    def __eq__(self, other):
        return self.fit is other.fit

    def __gt__(self, other):
        return self.fit > other.fit

    def __lt__(self, other):
        return self.fit < other.fit

    def __ne__(self, other):
        return self.fit is not other.fit

    def isdisjoint(self, other):
        return self.peak_indices.isdisjoint(other.peak_indices)

    def visit(self, other):
        if self.isdisjoint(other):
            self.edges.add(other)
            other.edges.add(self)
        else:
            self.overlap_edges.add(other)
            other.overlap_edges.add(self)

    def __repr__(self):
        return "FitNode(%r)" % self.fit


def tupleslices(iterable, shard_size=4):
    i = 0
    t = len(iterable)
    while i < t:
        yield iterable[i:i + shard_size]
        i += shard_size


def span_overlap(a, b):
    return a.__contains__(b.start) or a.__contains__(b.end - 1) or\
        b.__contains__(a.start) or b.__contains__(a.end - 1)


def peak_overlap(a, b):
    return len(a.peak_indices & b.peak_indices) > 0


def layout_layers(envelopes, overlap_fn=peak_overlap, maximize=True):
    """
    Produce a non-overlapping stacked layout of individual envelopes.
    """
    layers = [[]]
    envelopes.sort(key=lambda x: x.score, reverse=maximize)
    for envelope in envelopes:
        placed = False
        for layer in layers:
            collision = False
            for member in layer:
                if overlap_fn(envelope, member):
                    collision = True
                    break
            if not collision:
                layer.append(envelope)
                placed = True
                break
        if not placed:
            layers.append([envelope])
    return layers


class GreedySubgraphSelection(object):
    def __init__(self, subgraph, maximize=True, overlap_fn=peak_overlap):
        self.intervals = list(subgraph)
        self._layers = [[]]
        self.maximize = maximize
        self._build_layers(overlap_fn=overlap_fn)

    def _build_layers(self, overlap_fn):
        self._layers = layout_layers(self.intervals, overlap_fn=overlap_fn, maximize=self.maximize)

    def _select_best_subset(self):
        return self._layers[0]

    def select(self):
        return self._select_best_subset()

    @classmethod
    def solve(cls, nodes, maximize=True, overlap_fn=peak_overlap):
        solver = cls(nodes, maximize=maximize, overlap_fn=overlap_fn)
        solution = solver.select()
        return solution


class ExhaustiveDisjointSolutionSelection(object):  # pragma: no cover
    shard_size = 7

    def __init__(self, active_set, remaining_components, solutions=None, parent=None, score=None):
        if solutions is None:
            solutions = dict()
        if score is None:
            score = sum(f.score for f in active_set)
        self.active_set = active_set
        self.remaining_components = remaining_components
        self.solutions = solutions
        self.accepted = []
        self.rejected = []
        self.possible_next = []
        self.score = score

    def is_valid_addition(self, current, addition):
        is_valid = True
        for member in current:
            if not member.isdisjoint(addition):
                is_valid = False
                break
        return is_valid

    def check_has_solution(self, active_set, addition):
        for chunk in tupleslices(active_set, self.shard_size):
            try:
                check = self.solutions[chunk, addition.index]
            except KeyError:
                check = self.solutions[chunk, addition.index] = self.is_valid_addition(chunk, addition)
            if not check:
                return False
        return True

    def find_acceptable_peaks(self):
        accepted = []
        rejected = []
        for comp in self.remaining_components:
            if self.check_has_solution(self.active_set, comp):
                accepted.append(comp)
            else:
                rejected.append(comp)
        self.accepted = accepted
        self.rejected = rejected
        self.possible_next = self.remaining_components - set(rejected)

    def solve_level(self):
        if len(self.accepted) == 0:
            self.find_acceptable_peaks()
        accepted = self.accepted
        # next_steps = []
        current_step = set(self.active_set)
        extensions = self.possible_next
        for acc_ in accepted:
            acc = {acc_}
            next_set = tuple(sorted(current_step | acc))
            next_components = (extensions - acc) - acc_.overlaps
            new_score = self.score + acc_.score
            # total_possible_score = new_score + sum(f.score for f in next_components)
            # if total_possible_score > self.get_best_score() - 1:
            x = self.__class__(next_set, next_components, solutions=self.solutions,
                               parent=self, score=new_score)
            yield x

    def total_possible_score(self):
        if len(self.accepted) == 0:
            self.find_acceptable_peaks()
        if self.accepted:
            accepted_score = max(f.score for f in self.accepted)
        else:
            accepted_score = 0
        return self.score + sum(f.score for f in self.possible_next) + accepted_score

    @classmethod
    def solve(cls, nodes):
        initial_set = tuple()
        best_score = 0
        best_solution = None
        work_queue = GeneratorQueue()
        work_queue.extend(iter([cls(initial_set, set(nodes))]))
        i = 0
        # t = time.time()
        while work_queue.hasnext():
            i += 1
            solution = work_queue.next()
            solution.find_acceptable_peaks()

            if i % 300000 == 0 or solution.score > best_score and best_score != 0:
                # print i, best_score, len(best_solution.active_set), len(
                #     best_solution.remaining_components), len(
                #     best_solution.solutions), len(work_queue), best_solution.total_possible_score()
                # print '----', time.time() - t
                pass
            if solution.total_possible_score() < best_score:
                continue

            if solution.score > best_score:
                best_score = solution.score
                best_solution = solution

            work_queue.extend(solution.solve_level())
        return best_solution


class ConnectedSubgraph(object):
    def __init__(self, fits, maximize=True):
        self.nodes = tuple(map(FitNode, fits))
        self.maximize = maximize
        i = 0
        for node in self.nodes:
            node.index = i
            i += 1
        self.populate_edges()

    def __getitem__(self, i):
        return self.nodes[i]

    def __iter__(self):
        return iter(self.nodes)

    def __len__(self):
        return len(self.nodes)

    def populate_edges(self):
        for i in range(len(self.nodes)):
            node = self.nodes[i]
            for j in range(len(self.nodes)):
                if i == j:
                    continue
                other = self.nodes[j]
                node.visit(other)

    def find_heaviest_path(self, method="greedy"):
        if len(self) == 1:
            return set(self.nodes)
        if method == "greedy":
            solution = GreedySubgraphSelection.solve(tuple(self), maximize=self.maximize)
            return solution
        else:
            raise NotImplementedError(method)


try:
    has_c = True
    _ConnectedSubgraph = ConnectedSubgraph
    _FitNode = FitNode

    from ms_deisotope._c.peak_dependency_network.subgraph import ConnectedSubgraph, FitNode
except ImportError:
    has_c = False



"""Defines the base class for organizing the multiprocessing deconvolution algorithm
encapsulating all behaviors from start to finish.
"""
import os
import multiprocessing

from multiprocessing import JoinableQueue
from typing import Any, Dict, Iterator, List, Optional, Generic, TypeVar

from ms_deisotope.data_source.scan.loader import RandomAccessScanSource
from ms_deisotope.feature_map.scan_interval_tree import ScanIntervalTree

from ms_deisotope.processor import MSFileLoader

from ms_deisotope.feature_map.quick_index import index as build_scan_index
from ms_deisotope.task import TaskBase

from .collator import ScanCollator, T
from .process import ScanIDYieldingProcess, DeconvolutingScanTransformingProcess


class ScanGeneratorBase(Generic[T]):
    """The partial base class for the type that generates scan processing tasks.

    Defines getter and setters for properties expected to be used.

    Attributes
    ----------
    scan_source: :class:`~ms_deisotope.data_source.RandomAccessScanSource`
        Where to read scan data from.
    deconvoluting: bool
        Whether the transformation process will deisotope and charge state deconvolute.
    ms1_averaging: int
        The number of scans on either side of each MS1 scan to average with.
    extract_only_tandem_envelopes: bool
        Whether or not to process intervals around MSn precursors in MS1 scans or not.
    ignore_tandem_scans: bool
        Whether or not to ignore MSn scans.
    """
    scan_source: Optional[RandomAccessScanSource]

    ms1_averaging: int

    deconvoluting: bool
    extract_only_tandem_envelopes: bool
    ignore_tandem_scans: bool

    def configure_iteration(self, start_scan: Optional[int] = None, end_scan: Optional[int] = None, max_scans: Optional[int] = None):
        """Set :attr:`_iterator` to the result of :meth:`make_iterator`

        Parameters
        ----------
        start_scan : int, optional
            The starting scan index
        end_scan : int, optional
            The ending scan index
        max_scans : int, optional
            The maximum number of scans to process.
        """
        raise NotImplementedError()

    def make_iterator(self, start_scan: Optional[int] = None, end_scan: Optional[int] = None, max_scans: Optional[int] = None) -> Iterator[T]:
        """Create an :class:`Iterator` object over the scan product stream

        Parameters
        ----------
        start_scan : int, optional
            The starting scan index
        end_scan : int, optional
            The ending scan index
        max_scans : int, optional
            The maximum number of scans to process.

        Yields
        ------
        :class:`~.ProcessedScan`
        """
        raise NotImplementedError()

    _iterator = None

    def __iter__(self):
        return self

    def __next__(self) -> T:
        if self._iterator is None:
            self._iterator = self.make_iterator()
        return next(self._iterator)

    def next(self) -> T:
        return self.__next__()

    def close(self):
        pass

    @property
    def scan_source(self):
        return None

    _deconvoluting = False

    @property
    def deconvoluting(self):
        return self._deconvoluting

    @deconvoluting.setter
    def deconvoluting(self, value):
        self._deconvoluting = value

    _ms1_averaging = 0

    @property
    def ms1_averaging(self) -> int:
        return self._ms1_averaging

    @ms1_averaging.setter
    def ms1_averaging(self, value):
        self._ms1_averaging = value

    _ignore_tandem_scans = False

    @property
    def ignore_tandem_scans(self) -> bool:
        return self._ignore_tandem_scans

    @ignore_tandem_scans.setter
    def ignore_tandem_scans(self, value):
        self._ignore_tandem_scans = value

    _extract_only_tandem_envelopes = False

    @property
    def extract_only_tandem_envelopes(self) -> bool:
        return self._extract_only_tandem_envelopes

    @extract_only_tandem_envelopes.setter
    def extract_only_tandem_envelopes(self, value):
        self._extract_only_tandem_envelopes = value


class ScanGenerator(TaskBase, ScanGeneratorBase[T]):
    ms_file: os.PathLike
    verbose: bool

    number_of_helpers: int
    ms1_averaging: int

    scan_ids_exhausted_event: multiprocessing.Event
    default_precursor_ion_selection_window: float

    ms1_peak_picking_args: Dict[str, Any]
    msn_peak_picking_args: Dict[str, Any]

    ms1_deconvolution_args: Dict[str, Any]
    msn_deconvolution_args: Dict[str, Any]

    _scan_interval_tree: ScanIntervalTree
    _scan_yielder_process: ScanIDYieldingProcess
    _deconv_process: DeconvolutingScanTransformingProcess
    _deconv_helpers: List[DeconvolutingScanTransformingProcess]

    _order_manager: ScanCollator[T]

    _input_queue: multiprocessing.Queue
    _output_queue: multiprocessing.Queue


    def __init__(self, ms_file, number_of_helpers=4,
                 ms1_peak_picking_args=None, msn_peak_picking_args=None,
                 ms1_deconvolution_args=None, msn_deconvolution_args=None,
                 extract_only_tandem_envelopes=False, ignore_tandem_scans=False,
                 ms1_averaging=0, default_precursor_ion_selection_window=1.5,
                 deconvolute=True, verbose=False):
        self.ms_file = ms_file
        self.ignore_tandem_scans = ignore_tandem_scans

        self.scan_ids_exhausted_event = multiprocessing.Event()

        self._iterator = None

        self._scan_yielder_process = None
        self._deconv_process = None

        self._input_queue = None
        self._output_queue = None
        self._deconv_helpers = None
        self._order_manager = None

        self.number_of_helpers = number_of_helpers

        self.ms1_peak_picking_args = ms1_peak_picking_args
        self.msn_peak_picking_args = msn_peak_picking_args
        self.ms1_averaging = ms1_averaging
        self.default_precursor_ion_selection_window = default_precursor_ion_selection_window

        self.deconvoluting = deconvolute
        self.ms1_deconvolution_args = ms1_deconvolution_args
        self.msn_deconvolution_args = msn_deconvolution_args
        self.extract_only_tandem_envelopes = extract_only_tandem_envelopes
        self._scan_interval_tree = None
        self.verbose = verbose
        self.log_controller = self.ipc_logger()

    @property
    def scan_source(self):
        return self.ms_file

    def join(self):
        if self._scan_yielder_process is not None:
            self._scan_yielder_process.join()
        if self._deconv_process is not None:
            self._deconv_process.join()
        if self._deconv_helpers is not None:
            for helper in self._deconv_helpers:
                helper.join()

    def _terminate(self):
        if self._scan_yielder_process is not None:
            self._scan_yielder_process.terminate()
        if self._deconv_process is not None:
            self._deconv_process.terminate()
        if self._deconv_helpers is not None:
            for helper in self._deconv_helpers:
                helper.terminate()

    def _open_ms_file(self, **kwargs):
        return MSFileLoader(self.ms_file, **kwargs)

    def _preindex_file(self):
        reader = self._open_ms_file(use_index=False)
        try:
            reader.prebuild_byte_offset_file(self.ms_file)
        except AttributeError:
            # the type does not support this type of indexing
            pass
        except IOError:
            # the file could not be written
            pass
        except Exception as e:
            # something else went wrong
            self.error("An error occurred while pre-indexing.", e)

    def _make_interval_tree(self, start_scan: str, end_scan: str):
        reader = self._open_ms_file(decode_binary=False)
        if start_scan is not None:
            start_ix = reader.get_scan_by_id(start_scan).index
        else:
            start_ix = 0
        if end_scan is not None:
            end_ix = reader.get_scan_by_id(end_scan).index
        else:
            end_ix = len(reader)
        reader.reset()
        _index, interval_tree = build_scan_index(
            reader, self.number_of_helpers + 1, (start_ix, end_ix))
        self._scan_interval_tree = interval_tree

    def _make_transforming_process(self) -> DeconvolutingScanTransformingProcess:
        return DeconvolutingScanTransformingProcess(
            self.ms_file,
            self._input_queue,
            self._output_queue,
            self.scan_ids_exhausted_event,
            ms1_peak_picking_args=self.ms1_peak_picking_args,
            msn_peak_picking_args=self.msn_peak_picking_args,
            ms1_deconvolution_args=self.ms1_deconvolution_args,
            msn_deconvolution_args=self.msn_deconvolution_args,
            envelope_selector=self._scan_interval_tree,
            log_handler=self.log_controller.sender(),
            ms1_averaging=self.ms1_averaging,
            default_precursor_ion_selection_window=self.default_precursor_ion_selection_window,
            deconvolute=self.deconvoluting,
            verbose=self.verbose)

    def _make_collator(self) -> ScanCollator:
        return ScanCollator(
            self._output_queue, self.scan_ids_exhausted_event, self._deconv_helpers,
            self._deconv_process, input_queue=self._input_queue,
            include_fitted=not self.deconvoluting)

    def _make_scan_id_yielder(self, start_scan: str, end_scan: str, max_scans: int) -> ScanIDYieldingProcess:
        return ScanIDYieldingProcess(
            self.ms_file, self._input_queue, start_scan=start_scan, end_scan=end_scan,
            max_scans=max_scans, no_more_event=self.scan_ids_exhausted_event,
            ignore_tandem_scans=self.ignore_tandem_scans, batch_size=1,
            output_queue=self._output_queue)

    def _initialize_workers(self, start_scan: Optional[str] = None, end_scan: Optional[str] = None, max_scans: Optional[int] = None):
        try:
            self._input_queue = JoinableQueue(int(1e6))
            self._output_queue = JoinableQueue(int(1e6))
        except OSError:
            # Not all platforms permit limiting the size of queues
            self._input_queue = JoinableQueue()
            self._output_queue = JoinableQueue()

        self._preindex_file()

        if self.extract_only_tandem_envelopes:
            self.log("Constructing Scan Interval Tree")
            self._make_interval_tree(start_scan, end_scan)

        self._terminate()
        self._scan_yielder_process = self._make_scan_id_yielder(start_scan, end_scan, max_scans)
        self._scan_yielder_process.start()

        self._deconv_process = self._make_transforming_process()

        self._deconv_helpers = []

        for _ in range(self.number_of_helpers):
            self._deconv_helpers.append(self._make_transforming_process())
        self._deconv_process.start()

        self._order_manager = self._make_collator()

    def make_iterator(self, start_scan=None, end_scan=None, max_scans=None) -> Iterator[T]:
        self._initialize_workers(start_scan, end_scan, max_scans)

        for scan in self._order_manager:
            yield scan
        self.log_controller.stop()
        self.join()
        self._terminate()

    def configure_iteration(self, start_scan=None, end_scan=None, max_scans=None):
        self._iterator = self.make_iterator(start_scan, end_scan, max_scans)

    def close(self):
        self._terminate()



from __future__ import absolute_import

import json

from ms_deisotope.data_source import ChargeNotProvided

class JSONScanFormatter(object):

    def _pack_activation(self, activation_information):
        """Pack :class:`~.ActivationInformation` into a :class:`dict` structure
        which that :class:`~psims.mzml.writer.MzMLWriter` expects.

        Parameters
        ----------
        activation_information: :class:`~.ActivationInformation`

        Returns
        -------
        :class:`dict`
        """
        params = []
        params.append({
            "name": str(activation_information.method),
        })
        if activation_information.is_multiple_dissociation():
            for method in activation_information.methods[1:]:
                params.append({"name": str(method)})
        # NOTE: Only correct for CID/HCD spectra with absolute collision energies, but that is all I have
        # to test with.
        params.append({
            "name": "collision energy",
            "value": activation_information.energy,
            "unitName": "electronvolt"
        })
        if activation_information.is_multiple_dissociation():
            energies = activation_information.energies[1:]
            supplemental_energy = None
            if activation_information.has_supplemental_dissociation():
                supplemental_energy = energies[-1]
                energies = energies[:-1]
            for energy in energies:
                params.append({
                    "name": "collision energy",
                    "value": energy,
                    "unitName": "electronvolt"
                })
            if supplemental_energy is not None:
                params.append({
                    "name": 'supplemental collision energy',
                    "value": supplemental_energy,
                    "unitName": "electronvolt"
                })

        for key, val in activation_information.data.items():
            arg = {
                "name": key,
                "value": val
            }
            try:
                arg['unitName'] = val.unit_info
            except AttributeError:
                pass
            params.append(arg)
        return params

    def _pack_precursor_information(self, precursor_information, activation_information=None,
                                    isolation_window=None):
        """Repackage the :class:`~.PrecursorInformation`, :class:`~.ActivationInformation`,
        and :class:~.IsolationWindow` into the nested :class:`dict` structure that
        :class:`~psims.mzml.writer.MzMLWriter` expects.

        Parameters
        ----------
        precursor_information : :class:`~.PrecursorInformation`
        activation_information : :class:`~.ActivationInformation`, optional
        isolation_window : :class:`~.IsolationWindow`, optional

        Returns
        -------
        :class:`dict`
        """
        package = {}
        # If the scan bunch has been fully deconvoluted and it's PrecursorInformation
        # filled in, its extracted fields will be populated and should be used, otherwise
        # use the default read values.
        if precursor_information is not None:
            extracted_neutral_mass = precursor_information.extracted_neutral_mass
            if (extracted_neutral_mass != 0):
                package = {
                    "mz": precursor_information.extracted_mz,
                    "intensity": precursor_information.extracted_intensity,
                    "charge": precursor_information.extracted_charge,
                    "scan_id": precursor_information.precursor_scan_id,
                    "params": [
                        {"ms_deisotope:defaulted": precursor_information.defaulted},
                        {"ms_deisotope:orphan": precursor_information.orphan}
                    ]
                }
                if precursor_information.coisolation:
                    for p in precursor_information.coisolation:
                        package['params'].append({
                            "name": "ms_deisotope:coisolation",
                            "value": "%f %f %d" % (p.neutral_mass, p.intensity, p.charge)
                        })
            else:
                package = {
                    "mz": precursor_information.mz,
                    "intensity": precursor_information.intensity,
                    "charge": precursor_information.charge,
                    "scan_id": precursor_information.precursor_scan_id
                }
        else:
            package['mz'] = None
            package["charge"] = None
        if package['charge'] == ChargeNotProvided:
            package["charge"] = None
        if activation_information is not None:
            package['activation'] = self._pack_activation(
                activation_information)
        if isolation_window is not None:
            package['isolation_window'] = {
                "lower": isolation_window.lower,
                "target": isolation_window.target,
                "upper": isolation_window.upper
            }
        return package

    def peak_set_to_json(self, peak_set):
        points = []
        for peak in peak_set:
            points.append({
                "mz": peak.mz,
                "intensity": peak.intensity
            })
        return points

    def deconvoluted_peak_set_to_json(self, peak_set):
        points = []
        for peak in peak_set:
            points.append({
                "mz": peak.mz,
                "neutral_mass": peak.neutral_mass,
                "charge": peak.charge,
                "envelope": [
                    {'mz': p.mz, 'intensity': p.intensity} for p in peak.envelope
                ],
                "intensity": peak.intensity,
            })
        return points

    def raw_data_arrays_to_json(self, arrays):
        data = {
            "mz": arrays.mz.tolist(),
            "intensity": arrays.intensity.tolist()
        }
        return data

    def format_scan(self, scan):
        data = {}
        data['ms_level'] = scan.ms_level
        data['is_profile'] = scan.is_profile
        data['polarity'] = scan.polarity
        data['scan_time'] = scan.scan_time
        data['id'] = scan.id
        if scan.is_profile:
            data['arrays'] = self.raw_data_arrays_to_json(scan.arrays)
        if scan.peak_set is not None:
            data['peak_set'] = self.peak_set_to_json(scan.peak_set)
        if scan.deconvoluted_peak_set is not None:
            data['deconvoluted_peak_set'] = self.deconvoluted_peak_set_to_json(scan.deconvoluted_peak_set)
        if scan.precursor_information is not None:
            pinfo = data['precursor_information'] = self._pack_precursor_information(
                scan.precursor_information, scan.activation, scan.isolation_window)
            data['activation'] = pinfo.pop("activation", None)
            data['isolation_window'] = pinfo.pop("isolation_window")
        return data

    def dump(self, scan, fh, **kwargs):
        data = self.format_scan(scan)
        return json.dump(data, fh, **kwargs)

    def dumps(self, scan, **kwargs):
        data = self.format_scan(scan)
        return json.dumps(data, **kwargs)


formatter = JSONScanFormatter()


dump = formatter.dump
dumps = formatter.dumps
format = formatter.format_scan



from itertools import cycle
from collections import defaultdict
from typing import Optional

try:
    from matplotlib import pyplot as plt
    from matplotlib import patches as mpatches, colors as mcolors, collections as mcollections
    from matplotlib.colors import Normalize
except (RuntimeError, ImportError):
    pass
import numpy as np


from .profile_transform import interpolate, smooth_leveled


Ellipse = mpatches.Ellipse
FancyBboxPatch = mpatches.FancyBboxPatch
Rectangle = mpatches.Rectangle
LineCollection = mcollections.LineCollection


def compute_intensity_range(features):
    min_int = float('inf')
    max_int = 0
    for feature in features:
        for node in feature:
            t = node.total_intensity()
            min_int = min(min_int, t)
            max_int = max(max_int, t)
    return min_int, max_int


def feature_to_segments(feature):
    mzs = []
    times = []
    intensities = []
    for node in feature:
        mzs.append(node.mz)
        times.append(node.time)
        intensities.append(node.total_intensity())
    points = np.array([mzs, times]).T.reshape(-1, 1, 2)
    return np.concatenate([points[:-1], points[1:]], axis=1), np.array(intensities)


def draw_features(features, ax=None, alpha=0.65, norm=None, cmap=None, cbar=False, **kwargs):
    if norm is None:
        norm = Normalize(0, compute_intensity_range(features)[1])
    if ax is None:
        fig, ax = plt.subplots(1)
    if not features:
        return ax
    lines = []
    kwargs.setdefault("lw", 1)
    lw = kwargs.get("linewidth", kwargs.get("lw"))
    for feat in sorted(features, key=lambda x: x.intensity):
        segments, intensities = feature_to_segments(feat)
        art = LineCollection(segments, linewidths=lw, cmap=cmap, alpha=alpha, norm=norm)
        art.set_array(intensities)
        lines.append(art)

    for line in lines:
        ax.add_collection(line)

    if cbar:
        cbar = ax.figure.colorbar(lines[0])
        cbar.set_label("Intensity")

    ax.set_xlim(
        min(features, key=lambda x: x.mz if x is not None else float('inf')).mz - 1,
        max(features, key=lambda x: x.mz if x is not None else -float('inf')).mz + 1)
    ax.set_ylim(
        min(features, key=lambda x: x.start_time if x is not None else float(
            'inf')).start_time - 1,
        max(features, key=lambda x: x.end_time if x is not None else -float('inf')).end_time + 1)
    return ax


def draw_feature_sets(feature_sets, ax=None, alpha=0.65, width=2e-5, **kwargs):
    if ax is None:
        fig, ax = plt.subplots(1)

    kwargs.setdefault("lw", 0.05)
    lw = kwargs.get("linewidth", kwargs.get("lw"))

    features = []
    ellipse_sets = []
    for feature_set in feature_sets:
        ellipses = []
        for feat in feature_set:
            if feat is None:
                continue
            center = (feat.end_time + feat.start_time) / 2.
            height = feat.end_time - feat.start_time

            center_mz = feat.mz
            mz_width = center_mz * width

            ellipses.append(
                FancyBboxPatch((feat.mz - mz_width / 4., center - height / 2.),
                               width=mz_width / 2., height=height,
                               boxstyle=mpatches.BoxStyle.Round(pad=mz_width / 2.)))
            features.append(feat)
        ellipse_sets.append(ellipses)

    for ellipses in ellipse_sets:
        color = np.random.rand(3)
        for ell in ellipses:
            ell.set_alpha(alpha)
            ell.set_facecolor(color)
            ell.set_edgecolor(color)
            ell.set_linewidth(lw)
            ax.add_artist(ell)

    ax.set_xlim(
        min(features, key=lambda x: x.mz if x is not None else float('inf')).mz - 1,
        max(features, key=lambda x: x.mz if x is not None else -float('inf')).mz + 1)
    ax.set_ylim(
        min(features, key=lambda x: x.start_time if x is not None else float('inf')).start_time - 1,
        max(features, key=lambda x: x.end_time if x is not None else -float('inf')).end_time + 1)
    return ax


_nice_color_cycle = ([
    'blue', 'red', 'green', 'pink', 'orange', "grey", "purple",
    "seagreen", "darkgoldenrod", "darkcyan", "skyblue",
    "maroon", "darkgreen", "slategrey", "darkslateblue"
])


nice_color_cycle = cycle(_nice_color_cycle)


class Colorizer:

    def __call__(self, profile, *args, **kwargs):
        return 'blue'


class RandomColorizer(Colorizer):
    def __init__(self, seed: Optional[int]=42):
        super().__init__()
        self.rng = np.random.default_rng(seed)

    def __call__(self, profile, *args, **kwargs):
        return mcolors.rgb2hex(self.rng.random(3))


class NiceColorizer(Colorizer):
    def __init__(self, color_cycle=None):
        if color_cycle is None:
            color_cycle = _nice_color_cycle
        super().__init__()
        self.color_cycle = cycle(color_cycle)

    def __call__(self, profile, *args, **kwargs):
        return next(self.color_cycle)


random_colorizer = RandomColorizer()
nice_colorizer = NiceColorizer()


def labeler(profile, *args, **kwargs):
    label = "%0.4f" % profile.mz
    if hasattr(profile, 'charge'):
        label += ", %d" % profile.charge
    return label


def threshold_labeler(threshold):
    def label_func(profile, *args, **kwargs):
        if profile.intensity > threshold:
            return "%0.4f, %d" % (profile.neutral_mass, profile.charge)
        return ''
    return label_func


def draw_profiles(profiles, ax=None, smooth=False, interp=False, label_font_size=10,
                  axis_label_font_size=16, axis_font_size=16, label=True,
                  colorizer=None, label_function=labeler, legend=True):
    if ax is None:
        _fig, ax = plt.subplots(1)
    if colorizer is None:
        colorizer = RandomColorizer()
    if not profiles:
        return ax
    minimum_ident_time = float("inf")
    maximum_ident_time = 0
    maximum_intensity = 0

    _label_apexes = label and (label_function is not None)

    for profile in profiles:
        if profile is None:
            continue
        rt, heights = profile.as_arrays()

        if smooth:
            heights = smooth_leveled(rt, heights, smooth)
        if interp:
            rt, heights = interpolate(rt, heights)

        maximum_ident_time = max(max(rt), maximum_ident_time)
        minimum_ident_time = min(min(rt), minimum_ident_time)

        label = label_function(profile) if label_function is not None else None

        maximum_intensity = max(max(heights), maximum_intensity)
        color = colorizer(profile)
        ax.scatter(rt, heights, color=color)
        ax.fill_between(
            rt,
            heights,
            alpha=0.25,
            color=color,
            label=label
        )
        apex = max(heights)
        apex_ind = np.argmax(heights)
        rt_apex = rt[apex_ind]
        if _label_apexes:
            ax.text(rt_apex, apex + min(apex * (
                .1), 1200), label,
                ha='center', fontsize=label_font_size, alpha=0.75)
    ax.set_xlim(minimum_ident_time - 0.02,
                maximum_ident_time + 0.02)
    ax.set_ylim(0, maximum_intensity * 1.1)
    if legend:
        ax.legend(loc='upper left', bbox_to_anchor=(1.05, 1.), ncol=2, fontsize=10)
    ax.axes.spines['right'].set_visible(False)
    ax.axes.spines['top'].set_visible(False)
    ax.yaxis.tick_left()
    ax.xaxis.tick_bottom()
    ax.set_xlabel("Time", fontsize=axis_label_font_size)
    ax.set_ylabel("Relative Abundance", fontsize=axis_label_font_size)
    [t.set(fontsize=axis_font_size) for t in ax.get_xticklabels()]
    [t.set(fontsize=axis_font_size) for t in ax.get_yticklabels()]
    return ax


def features_to_peak_time_pairs(features):
    pairs = []
    for feature in features:
        for node in feature:
            for peak in node.members:
                pairs.append((peak, node.time))
    pairs.sort(key=lambda x: x[1])
    return pairs


def extract_intensity_array(peaks):
    mzs = []
    intensities = []
    rts = []
    current_mzs = []
    current_intensities = []
    last_time = None
    time = None
    for peak, time in peaks:
        if time != last_time:
            if last_time is not None:
                mzs.append(current_mzs)
                intensities.append(current_intensities)
                rts.append(time)
            last_time = time
            current_mzs = []
            current_intensities = []
        current_mzs.append(peak.mz)
        current_intensities.append(peak.intensity)
    if time is not None:
        mzs.append(current_mzs)
        intensities.append(current_intensities)
        rts.append(time)
    return mzs, intensities, rts


def binner(x):
    return np.floor(np.array(x) / 10.) * 10


def make_map(mzs, intensities):
    binned_mzs = [
        binner(mz_row) for mz_row in mzs
    ]
    unique_mzs = set()
    list(map(unique_mzs.update, binned_mzs))
    unique_mzs = np.array(sorted(unique_mzs))
    assigned_bins = []
    j = 0
    for mz, inten in zip(mzs, intensities):
        j += 1
        mz = binner(mz)
        bin_row = defaultdict(float)
        for i in range(len(mz)):
            k = mz[i]
            v = inten[i]
            bin_row[k] += v
        array_row = np.array([bin_row[m] for m in unique_mzs])
        assigned_bins.append(array_row)

    assigned_bins = np.vstack(assigned_bins)
    assigned_bins[assigned_bins == 0] = 1
    return assigned_bins, unique_mzs


def render_map(assigned_bins, rts, unique_mzs, ax=None, color_map=None, scaler=np.sqrt, colorbar=True):
    if ax is None:
        _fig, ax = plt.subplots(1)
    if color_map is None:
        color_map = plt.cm.viridis
    data = np.array(scaler(assigned_bins.T))
    img = ax.pcolormesh(data, cmap=color_map)

    xticks = ax.get_xticks()
    newlabels = np.array(ax.get_xticklabels())[np.arange(0, len(xticks))]
    n = len(newlabels)
    interp = np.linspace(rts[0], rts[-1], n)
    for i, label in enumerate(newlabels):
        num = round(interp[i], 1)
        label.set_rotation(90)
        label.set_text(str((num)))

    # ax.set_xticklabels(newlabels)

    yticks = ax.get_yticks()
    n = len(yticks)
    # print n, yticks
    newlabels = np.array(ax.get_yticklabels())[np.arange(0, len(yticks))]

    va = unique_mzs
    # Hacky and doesn't align quite right
    n = len(newlabels)
    # if n % 2 != 0:
    #     n -= 1
    interp = np.linspace(va[0], va[-1], n)
    newlabels = []
    # for i, label in enumerate(newlabels):
    for i in range(n):
        num = interp[i]
        newlabels.append(str(round(num, 2)))
    ax.set_yticklabels((newlabels))
    ax.set_xlabel("Time")
    ax.set_ylabel("m/z")
    if colorbar:
        _cbar = ax.figure.colorbar(img, ax=ax)
    return ax


def heatmap(features, ax=None, color_map=None, scaler=np.sqrt, colorbar=True):
    mzs, intensities, rts = extract_intensity_array(
        features_to_peak_time_pairs(features))
    binned_intensities, unique_mzs = make_map(mzs, intensities)
    return render_map(
        binned_intensities, rts, unique_mzs, ax=ax, color_map=color_map,
        scaler=scaler, colorbar=colorbar)



"""
    Waters
    MassLynx Python SDK
"""

import copy
import ctypes
from ctypes import*
from array import *
from .MassLynxRawReader import *
from .MassLynxRawDefs import MassLynxBaseType

class MassLynxLockMassProcessor(object):
    """Class to enable post acquisition lock mass correction"""

    def __init__(self):
        self._codeHandler = MassLynxCodeHandler()
        self.mlLockMassProcessor = c_void_p()
        createRawProcessor = MassLynxRawReader.massLynxDll.createRawProcessor
        createRawProcessor.argtypes = [POINTER(c_void_p), c_int, c_void_p, POINTER(c_void_p)]
        self._codeHandler.CheckReturnCode(createRawProcessor(self.mlLockMassProcessor, MassLynxBaseType.LOCKMASS, c_void_p(0), c_void_p(0) ))

    # destroy the processor
    def __del__(self):
        destroyLockMassProcessor = MassLynxRawReader.massLynxDll.destroyRawProcessor
        destroyLockMassProcessor.argtypes = [c_void_p]
        destroyLockMassProcessor( self.mlLockMassProcessor )

    def SetRawData( self, source ):
        # set data from a path from a path
        if (isinstance(source, str) ):
            self._setRawPath(source)

        # set data from a reader
        elif (isinstance(source, MassLynxRawReader)):
            self._setRawReader(source)

        # did we fall through
        else:
            self._codeHandler.CheckReturnCode( 1 )

        return

    def _setRawPath( self, path ):
        bytes = str.encode(path)
        setRawPath = MassLynxRawReader.massLynxDll.setRawPath
        setRawPath.argtypes = [c_void_p, c_char_p]
        self._codeHandler.CheckReturnCode(setRawPath(self.mlLockMassProcessor, bytes) )


    def _setRawReader( self, mlReader ):
        setRawReader = MassLynxRawReader.massLynxDll.setRawReader
        setRawReader.argtypes = [c_void_p, c_void_p]
        self._codeHandler.CheckReturnCode(setRawReader(self.mlLockMassProcessor, mlReader._getReader() ))

    def SetParameters(self, parameters):
        p_params = parameters.GetParameters()
        setLockMassParameters = MassLynxRawReader.massLynxDll.setLockMassParameters
        setLockMassParameters.argtypes = [c_void_p, c_void_p]
        self._codeHandler.CheckReturnCode(
            setLockMassParameters(self.mlLockMassProcessor, p_params))

    def LockMassCorrect( self ):
         # get applied lock mass gain
        success = c_bool(0)
        lockMassCorrect =  MassLynxRawReader.massLynxDll.lockMassCorrect
        lockMassCorrect.argtypes = [c_void_p, POINTER(c_bool)]
        self._codeHandler.CheckReturnCode(lockMassCorrect(self.mlLockMassProcessor, success ))

        return success.value

    def RemoveLockMassCorrection( self ):
        # get applied lock mass gain
        removeLockMassCorrection =  MassLynxRawReader.massLynxDll.removeLockMassCorrection
        removeLockMassCorrection.argtypes = [c_void_p]
        self._codeHandler.CheckReturnCode( removeLockMassCorrection(self.mlLockMassProcessor) )

    def IsLockMassCorrected( self ):
        # get applied lock mass gain
        applied = c_int(0)
        isLockMassCorrected =  MassLynxRawReader.massLynxDll.LMP_isLockMassCorrected
        isLockMassCorrected.argtypes = [c_void_p, POINTER(c_int)]
        self._codeHandler.CheckReturnCode( isLockMassCorrected(self.mlLockMassProcessor, applied) )

        return applied.value == 1

    def CanLockMassCorrect( self ):
        # get applied lock mass gain
        canApply = c_int(0)
        canLockMassCorrect =  MassLynxRawReader.massLynxDll.LMP_canLockMassCorrect
        canLockMassCorrect.argtypes = [c_void_p, POINTER(c_int)]
        self._codeHandler.CheckReturnCode( canLockMassCorrect(self.mlLockMassProcessor, canApply) )

        return canApply.value == 1

    def GetLockMassValues( self ):
        # get applied lock mass gain
        mass = c_float(0)
        tolerance = c_float(0)
        getLockMassValues =  MassLynxRawReader.massLynxDll.getLockMassValues
        getLockMassValues.argtypes = [c_void_p, POINTER(c_float), POINTER(c_float)]
        self._codeHandler.CheckReturnCode( getLockMassValues(self.mlLockMassProcessor, mass, tolerance) )

        return mass.value, tolerance.value

    def GetLockMassCorrection( self, retentionTime ):
        # get applied lock mass gain
        gain = c_float(0)
        getLockMassCorrection =  MassLynxRawReader.massLynxDll.getLockMassCorrection
        getLockMassCorrection.argtypes = [c_void_p, c_float, POINTER(c_float)]
        self._codeHandler.CheckReturnCode( getLockMassCorrection(self.mlLockMassProcessor, retentionTime, gain) )

        return gain.value



"""A collection of simple statistics describing isotopic envelopes.
These functions are primarily used during deconvolution to calculate
properties of :class:`~.DeconvolutedPeak` objects.
"""
import operator

intensity_getter = operator.attrgetter("intensity")
mz_getter = operator.attrgetter("mz")
snr_getter = operator.attrgetter("signal_to_noise")


def a_to_a2_ratio(envelope):
    """Calculate the ratio of the abundance of the A+0 (monoisotopic)
    peak to the abundance of the A+2 (+2 neutron) peak.

    Parameters
    ----------
    envelope : :class:`list` of :class:`~.FittedPeak`
        The sequence of experimental peaks matched.

    Returns
    -------
    float
    """
    if len(envelope) < 3:
        return 0.
    a0 = envelope[0]
    a2 = envelope[2]
    if a0.mz < 0 or a2.mz < 0:
        return 0.
    return a0.intensity / a2.intensity


def total_area(envelope):
    """Sum the area of all peaks in the envelope.

    Parameters
    ----------
    envelope : :class:`list` of :class:`~.FittedPeak`
        The sequence of experimental peaks matched.

    Returns
    -------
    float
    """
    return sum(p.area for p in envelope)


def most_abundant_mz(envelope):
    """Find the most abundant m/z peak in the envelope

    Parameters
    ----------
    envelope : :class:`list` of :class:`~.FittedPeak`
        The sequence of experimental peaks matched.

    Returns
    -------
    float
    """
    return max([p for p in envelope if p.mz > 1], key=intensity_getter).mz


def average_mz(envelope):
    """An intensity weighted average the m/z of the peaks
    in the envelope

    Parameters
    ----------
    envelope : :class:`list` of :class:`~.FittedPeak`
        The sequence of experimental peaks matched.

    Returns
    -------
    float
    """
    envelope = [p for p in envelope if p.mz > 1]
    return weighted_average(map(mz_getter, envelope), map(intensity_getter, envelope))


def average_signal_to_noise(envelope):
    """Average the signal to noise ratio of the peaks in the envelope

    Parameters
    ----------
    envelope : :class:`list` of :class:`~.FittedPeak`
        The sequence of experimental peaks matched.

    Returns
    -------
    float
    """
    envelope = [p for p in envelope if p.mz > 1]
    return sum(map(snr_getter, envelope)) / float(len(envelope))


def weighted_average(values, weights):
    """Calculate a weighted average over `values` weighted by `weights`.

    Parameters
    ----------
    values : :class:`~.Iterable`
        The values
    weights : :class:`~.Iterable`
        The weights

    Returns
    -------
    float
    """
    # We need to traverse weights twice, therefore convert it to a list
    # as a generator (i.e. py3's map) would always have sum(weights) == 0
    weights = list(weights)
    return sum(v * w for v, w in zip(values, weights)) / float(sum(weights))



""" Waters
    MassLynx Python SDK
"""

import string
import ctypes

from ctypes import *
from .libload import proxy, register_dll

#from enum import IntEnum

#class MassLynxScanItem(IntEnum):
#    LINEAR_DETECTOR_VOLTAGE = 1
#    LINEAR_SENSITIVITY = 2
#    REFLECTRON_LENS_VOLTAGE = 3
#    REFLECTRON_DETECTOR_VOLTAGE = 4
#    REFLECTRON_SENSITIVITY = 5
#    LASER_REPETITION_RATE = 6
#    COURSE_LASER_CONTROL = 7
#    FINE_LASER_CONTROL = 8
#    LASERAIM_XPOS = 9
#    LASERAIM_YPOS = 10
#    NUM_SHOTS_SUMMED = 11
#    NUM_SHOTS_PERFORMED = 12
#    SEGMENT_NUMBER = 13
#    LCMP_TFM_WELL = 14
#    SEGMENT_TYPE = 15
#    SOURCE_REGION1 = 16
#    SOURCE_REGION2 = 17
#    REFLECTRON_FIELD_LENGTH = 18
#    REFLECTRON_LENGTH = 19
#    REFLECTRON_VOLT = 20
#    SAMPLE_PLATE_VOLT = 21
#    REFLECTRON_FIELD_LENGTH_ALT = 22
#    REFLECTRON_LENGTH_ALT = 23
#    PSD_STEP_MAJOR = 24
#    PSD_STEP_MINOR = 25
#    PSD_FACTOR_1 = 26
#    NEEDLE = 50
#    COUNTER_ELECTRODE_VOLTAGE = 51
#    SAMPLING_CONE_VOLTAGE = 52
#    SKIMMER_LENS = 53
#    SKIMMER = 54
#    PROBE_TEMPERATURE = 55
#    SOURCE_TEMPERATURE = 56
#    RF_VOLTAGE = 57
#    SOURCE_APERTURE = 58
#    SOURCE_CODE = 59
#    LM_RESOLUTION = 60
#    HM_RESOLUTION = 61
#    COLLISION_ENERGY = 62
#    ION_ENERGY = 63
#    MULTIPLIER1 = 64
#    MULTIPLIER2 = 65
#    TRANSPORTDC = 66
#    TOF_APERTURE = 67
#    ACC_VOLTAGE = 68
#    STEERING = 69
#    FOCUS = 70
#    ENTRANCE = 71
#    GUARD = 72
#    TOF = 73
#    REFLECTRON = 74
#    COLLISION_RF = 75
#    TRANSPORT_RF = 76
#    SET_MASS = 77
#    COLLISION_ENERGY2 = 78
#    SONAR_ENABLED = 81
#    QUAD_START_MASS = 82
#    QUAD_STOP_MASS = 83
#    QUAD_PEAK_WIDTH = 84
#    REFERENCE_SCAN = 100
#    USE_LOCKMASS_CORRECTION = 101
#    LOCKMASS_CORRECTION = 102
#    USETEMP_CORRECTION = 103
#    TEMP_CORRECTION = 104
#    TEMP_COEFFICIENT = 105
#    FAIMS_COMPENSATION_VOLTAGE = 106
#    TIC_TRACE_A = 107
#    TIC_TRACE_B = 108
#    RAW_EE_CV = 110
#    RAW_EE_CE = 111
#    ACCURATE_MASS = 112
#    ACCURATE_MASS_FLAGS = 113
#    SCAN_ERROR_FLAG = 114
#    DRE_TRANSMISSION = 115
#    SCAN_PUSH_COUNT = 116
#    RAW_STAT_SWAVE_NORMALISATION_FACTOR = 117
#    MIN_DRIFT_TIME_CHANNEL = 122
#    MAX_DRIFT_TIME_CHANNEL = 123
#    TOTAL_ION_CURRENT = 252
#    BASE_PEAK_MASS = 253
#    BASE_PEAK_INTENSITY = 254
#    PEAKS_IN_SCAN = 255
#    UNINITIALISED = 1000


class MassLynxException(Exception):
    code = 0
    def __init__(self, code, message):
        self.message = message
        self.code = code
    def __str__(self):
        return repr(  self.message )

    def Handler(self):
        #print( self.message )
        # can rethrow if needed
        return

# string handler
class MassLynxStringHandler(object):

    def __init__(self):
        return

    def ToString(self, chString, release):
        if (None == chString):
            return ""

        strValue  = chString.value.decode()
        if (release ):
            MassLynxRawReader.ReleaseMemory(chString)

        chString = None

        return strValue


class MassLynxCodeHandler(object):
    def __init__(self):
        self._code = 0
        self._stringHandler = MassLynxStringHandler()
        return

    # three option true, false, throw exception
    def CheckReturnCode( self, code, throw = True ):
        self._code = code;
        if (0 == code):
            return True

        if (throw):
            raise MassLynxException( self.GetLastCode(), self.GetLastMessage() )

        # get last error
        return False

    def GetLastCode(self):
        return self._code

    def GetLastMessage(self):
        # load the dll
        getErrorMessage = MassLynxRawReader.massLynxDll.getErrorMessage
        getErrorMessage.argtypes = [ c_int, POINTER(c_char_p)]

        message = (c_char_p)()
        getErrorMessage( self.GetLastCode(), message )
  #      exceptionMessage = "MassLynx Exception {} : {}".format( returnCode, message.value.decode())

        # release the memory
        return self._stringHandler.ToString(message, True)



class MassLynxRawReader(object):
    """basic functionality to read raw files"""

    # load the dll
    massLynxDll = proxy
    version = '1.0'                                 # class variable

    def __init__(self, source, mlType):

        self.mlRawReader = c_void_p()              # instance variable
        self._codeHandler = MassLynxCodeHandler()
        self._stringHandler = MassLynxStringHandler()

        # create scan reader from a path
        if (isinstance(source, str) ):
            bytes = str.encode(source)
            createRawReaderFromPath = MassLynxRawReader.massLynxDll.createRawReaderFromPath
            createRawReaderFromPath.argtypes = [ c_char_p, POINTER(c_void_p), c_int]
            self._codeHandler.CheckReturnCode(createRawReaderFromPath(bytes, self._getReader(), mlType))

        # create scan reader from a reader
        elif (isinstance(source, MassLynxRawReader)):
            createRawReaderFromReader = MassLynxRawReader.massLynxDll.createRawReaderFromReader
            createRawReaderFromReader.argtypes = [ c_void_p, POINTER(c_void_p), c_int]
            self._codeHandler.CheckReturnCode(createRawReaderFromReader(source._getReader(),self._getReader(),mlType))

        # did we fall through
        else:
            self._codeHandler.CheckReturnCode( 1 )

        return

    # destroy the reader
    def __del__(self):
        # destroy reader
        destroyRawReader = MassLynxRawReader.massLynxDll.destroyRawReader
        destroyRawReader.argtypes = [c_void_p]
        destroyRawReader( self._getReader() )

    def _getReader(self):
        return self.mlRawReader

    def ToString( self, pString, release = False ):
        return self._stringHandler.ToString( pString, release )

    def CheckReturnCode(self, code, throw = True):
        return self._codeHandler.CheckReturnCode( code , throw )

    # common util to free memory
    @staticmethod
    def ReleaseMemory( address):
        releaseMemory = MassLynxRawReader.massLynxDll.releaseMemory
        releaseMemory.argtypes = [ c_void_p]
        releaseMemory( address )


"""Simple library-wide configuration management to handle
the tracking of external 3rd-party libraries and machine state.

"""
import os
import sys
import copy
import json
import warnings

from six import PY2


CONFIG_FILE_NAME = 'config.json'


def _get_home_dir():
    """Find user's home directory if possible.
    Otherwise, returns None.

    :see:
        http://mail.python.org/pipermail/python-list/2005-February/325395.html
    """
    if PY2 and sys.platform == 'win32':
        path = os.path.expanduser(b"~").decode(sys.getfilesystemencoding())
    else:
        path = os.path.expanduser("~")
    if os.path.isdir(path):
        return path
    for evar in ('HOME', 'USERPROFILE', 'TMP'):
        path = os.environ.get(evar)
        if path is not None and os.path.isdir(path):
            return path
    return None


def get_config_dir():
    """Get the configuration directory path.

    Tries the following routes:
        1. The environment variable "MS_DEISOTOPE_CONFIGDIR"
        2. If on an XDG-compliant platform (Linux, Free BSD), the environment
           variable "XDG_CONFIG_HOME"/ms_deisotope
        3. The user's home directory/.ms_deisotope

    If the configuration directory does not exist, it will
    be created.

    Returns
    -------
    str
    """
    FALLBACK_DIR = '.'
    configdir = os.environ.get('MS_DEISOTOPE_CONFIGDIR')
    if configdir is not None:
        configdir = os.path.abspath(configdir)
        if not os.path.exists(configdir):
            try:
                os.makedirs(configdir)
            except OSError:
                return FALLBACK_DIR
        return configdir

    home_dir = _get_home_dir()
    if home_dir is not None:
        configdir = os.path.join(home_dir, '.ms_deisotope')
    else:
        configdir = None
    if sys.platform.startswith(('linux', 'freebsd')):
        configdir = None
        xdg_base = os.environ.get('XDG_CONFIG_HOME')
        if xdg_base is None:
            xdg_base = home_dir
            if xdg_base is not None:
                xdg_base = os.path.join(xdg_base, '.config')
        if xdg_base is not None:
            configdir = os.path.join(xdg_base, 'ms_deisotope')
    if configdir is not None:
        if os.path.exists(configdir):
            return configdir
        try:
            os.makedirs(configdir)
            return configdir
        except OSError:
            return FALLBACK_DIR
    return FALLBACK_DIR


_DEFAULT_CONFIG = {
    'vendor_readers': {
        "thermo-com": [],
        "thermo-net": [],
        "agilent-com": [],
        "waters-masslynx": [],
    },
    'schema_version': '1.0.0'
}


def get_config():
    """Load the config.json file from the configuration
    directory given by :func:`get_config_dir`.

    If the config file does not exist, a default one
    will be created.

    Returns
    -------
    dict
    """
    confdir = get_config_dir()
    path = os.path.join(confdir, CONFIG_FILE_NAME)
    if not os.path.exists(path):
        try:
            save_config(_DEFAULT_CONFIG)
        except OSError as err:
            warnings.warn("Encountered the error %r when trying to write the default configuration" % (err, ))
            return copy.deepcopy(_DEFAULT_CONFIG)
    try:
        with open(path, 'rt') as fh:
            config = json.load(fh)
    except OSError as err:
        warnings.warn(
            "Encountered the error %r when trying to read configuration" % (err, ))
        config = copy.deepcopy(_DEFAULT_CONFIG)
    if "schema_version" not in config:
        config['schema_version'] = '1.0.0'
    return config


def save_config(config=None):
    """Save the configuration in `config` to disk in the
    configuration directory given by :func:`get_config_dir`.

    Parameters
    ----------
    config : dict, optional
        The configuration dictionary (the default is None, which will use the default, empty config)

    """
    if config is None:
        config = copy.deepcopy(_DEFAULT_CONFIG)
    confdir = get_config_dir()
    path = os.path.join(confdir, CONFIG_FILE_NAME)
    with open(path, 'wt') as fh:
        json.dump(config, fh)



import os
import re
import time
import logging
import hashlib
import json

import dataclasses
from dataclasses import dataclass, field as dfield
from collections import defaultdict
from threading import RLock, Thread, local
from typing import Dict, List

import idzip

import ms_deisotope
from ms_deisotope.data_source.scan.base import ScanBase

from ms_deisotope.utils import LRUDict
from ms_deisotope.data_source._compression import get_opener
from ms_deisotope.data_source.usi import USI


BLOCK_SIZE = 2 ** 20
logger = logging.getLogger("proxi_backend")

@dataclass(frozen=True)
class FileType:
    accession: str
    name: str

    def __str__(self):
        return self.name

    def to_dict(self) -> Dict:
        return dataclasses.asdict(self)


PeakListFile = FileType("MS:1002850", "Peak list file URI")
AssociatedRawFile = FileType("MS:1002846", "Associated raw file URI")


@dataclass(frozen=True)
class PROXIDataFile:
    uri: str
    file_type: FileType

    def to_dict(self) -> Dict:
        d = self.file_type.to_dict()
        d['value'] = self.uri
        return d

    @classmethod
    def from_dict(cls, d):
        ft = FileType(d['accession'], d['name'])
        return cls(d['value'], ft)


@dataclass
class PROXIDataset:
    accession: str = dfield(default=None)
    title: str = dfield(default=None)
    summary: str = dfield(default=None)
    instruments: List[str] = dfield(default_factory=list)
    identifiers: List[str] = dfield(default_factory=list)
    species: List[str] = dfield(default_factory=list)
    modifications: List[str] = dfield(default_factory=list)
    contacts: List[str] = dfield(default_factory=list)
    publications: List[str] = dfield(default_factory=list)
    keywords: List[str] = dfield(default_factory=list)
    dataset_link: str = dfield(default=None)
    dataset_files: List[PROXIDataFile] = dfield(default_factory=list)
    links: List[str] = dfield(default_factory=list)

    def to_dict(self):
        d = dataclasses.asdict(self)
        d['datasetLinks'] = d.pop("dataset_links", None)
        # TODO: Make this consistent whether or not it is a dict or an instance
        d['dataFiles'] = [PROXIDataFile(f['uri'], FileType(**f['file_type'])).to_dict() for f in d.pop('dataset_files', [])]
        return d

    @classmethod
    def from_dict(cls, d):
        d = d.copy()
        d['dataset_links'] = d.pop('datasetLinks', None)
        d['dataset_files'] = [
            PROXIDataFile.from_dict(f) for f in d.pop('datasetFiles', [])]
        return cls(**d)

    @classmethod
    def new(cls, dataset_id):
        self = cls()
        self.accession = dataset_id
        self.title = dataset_id
        self.summary = dataset_id
        return self


class PROXIDatasetAPIMixinBase(object):

    def _datafile_uri(self, dataset_id, data_file):
        raise NotImplementedError()

    def _is_peaklist(self, d):
        base = os.path.basename(d)
        if base.endswith(".gz"):
            base = base[:-3]
        base = base.lower()
        _, ext = base.rsplit(".", 1)
        if ext in tuple(map(str.lower, self.peaklist_file_types)):
            return True
        return False

    @property
    def peaklist_file_types(self):
        """Alias to allow :attr:`valid_file_types` to vary.

        Returns
        -------
        list[str]"""
        return self.valid_file_types

    def describe_dataset(self, dataset_id: str) -> PROXIDataset:
        record = PROXIDataset.new(dataset_id)
        data_files = self._find_mz_files(self.index[dataset_id])
        for d in data_files:
            uri = self._datafile_uri(dataset_id, d)
            if self._is_peaklist(d):
                f = PROXIDataFile(uri, PeakListFile)
            else:
                f = PROXIDataFile(uri, AssociatedRawFile)
            record.dataset_files.append(f)
        return record

    def enumerate_datasets(self, result_type=None):
        if result_type == 'compact':
            result = list(self.index.keys())
            return result
        return [
            self.describe_dataset(d).to_dict() for d in self.index.keys()]


class IndexManagingMixin(object):

    def __init__(self, **kwargs):
        self.index = self.build_index()
        self.monitor_thread = self.monitor_index(
            kwargs.get("update_frequency"))

    def _monitor_index(self, update_frequency=None):
        if update_frequency is None:
            # Default to one hour, 60^2 seconds
            update_frequency = 60.0 ** 2
        logger.info(
            "Monitoring archive for changes every %0.2f seconds.", update_frequency)
        while True:
            time.sleep(update_frequency)
            logger.info("Rebuilding Index!")
            self.index = self.build_index(use_cache=False)

    def monitor_index(self, update_frequency=None):
        update_thread = Thread(
            target=self._monitor_index,
            args=(update_frequency, ))
        update_thread.daemon = True
        update_thread.start()
        return update_thread

    def _checksum(self, string):
        return hashlib.new("md5", string).hexdigest()

    def _find_index_cache(self, use_cache=True):
        cache_name = hashlib.md5(
            self.base_uri.encode('utf8')).hexdigest() + '.json'
        if os.path.exists(cache_name) and use_cache:
            logger.info("Reading cache from %r", cache_name)
            try:
                with open(cache_name, 'rt') as fh:
                    index = defaultdict(list, json.load(fh))
                    return cache_name, index
            except IOError:
                logger.error("Failed to read index from disk", exc_info=True)
        return cache_name, None


class SpectrumSerializerServiceMixin(object):

    def build_response(self, usi: USI, scan: ScanBase):
        payload = {
            "usi": str(usi),
            "status": "READABLE",
            "attributes": [
                {"accession": "MS:1000511", "name": "ms level",
                    "value": scan.ms_level},
                {"accession": "MS:1008025", "name": "scan number",
                    "value": scan.index + 1},
                {"accession": "MS:1003061", "name": "spectrum name", "value": scan.id},
            ]
        }
        payload = self._write_peak_arrays(usi, scan, payload)
        if scan.ms_level > 1:
            mz = None
            charge = None
            if scan.precursor_information:
                mz = scan.precursor_information.mz
                charge = scan.precursor_information.charge
            elif scan.isolation_window:
                mz = scan.isolation_window.target
            if mz:
                payload['attributes'].append({
                    "accession": "MS:1000827", "name": "isolation window target m/z",
                    "value": mz,
                })
            if charge:
                payload['attributes'].append({
                    "accession": "MS:1000041", "name": "charge state",
                    "value": charge,
                })
        return payload

    def _write_peak_arrays(self, usi, scan, payload):
        mzs = []
        intensities = []
        charges = None
        base_peak = None
        nb_peaks = None
        if scan.deconvoluted_peak_set is not None:
            mzs = [p.mz for p in scan.deconvoluted_peak_set]
            intensities = [p.intensity for p in scan.deconvoluted_peak_set]
            charges = [p.charge for p in scan.deconvoluted_peak_set]
            is_profile = False
            nb_peaks = len(scan.deconvoluted_peak_set)
            base_peak = scan.base_peak.deconvoluted().intensity
        elif scan.peak_set is not None:
            mzs = [p.mz for p in scan.peak_set]
            intensities = [p.intensity for p in scan.peak_set]
            charges = None
            is_profile = False
            nb_peaks = len(scan.peak_set)
            base_peak = scan.base_peak.centroided().intensity
        else:
            mzs = scan.arrays.mz.tolist()
            intensities = scan.arrays.intensity.tolist()
            is_profile = scan.is_profile
            nb_peaks = len(mzs)
            base_peak = scan.base_peak.raw().intensity

        payload['m/z array'] = mzs
        payload['intensity array'] = intensities
        if charges is not None:
            payload['charge array'] = charges
        payload['attributes'].append(
            {"accession": "MS:1000128", "name": "profile spectrum"} if is_profile else
            {"accession": "MS:1000127", "name": "centroid spectrum"})
        payload['attributes'].append({
            "accession": 'MS:1000505', 'name': 'base peak intensity', 'value': base_peak
        })
        payload['attributes'].append({
            "accession": 'MS:1003059', 'name': 'number of peaks', 'value': nb_peaks
        })
        return payload


class DatasetNotAvailable(ValueError):
    pass


class UnrecognizedIndexFlag(ValueError):
    pass


class MassSpectraDataArchiveBase(SpectrumSerializerServiceMixin, PROXIDatasetAPIMixinBase, IndexManagingMixin):
    valid_file_types = ['mzML', 'mzXML', 'mgf', 'mzMLb']
    special_openers = set()

    def __init__(self, base_uri, cache_size=None, **kwargs):
        if cache_size is None:
            cache_size = 1
        self.base_uri = base_uri
        self.cache = LRUDict(cache_size=cache_size)
        IndexManagingMixin.__init__(self, **kwargs)

    def build_index(self, use_cache=True):
        file_types = self.valid_file_types + ['json']

        cache_name, index = self._find_index_cache(use_cache=use_cache)
        if index is not None and use_cache:
            return index

        index = defaultdict(list)

        self._traverse_files(file_types, index)

        reduced = {
            k: self._find_mz_files(v) for k, v in index.items()
        }

        n_datasets = len(reduced)
        n_ms_files = 0
        for _, v in reduced.items():
            n_ms_files += len(v)

        logger.info(
            "Collected %d Datasets with %d MS Data Files in Index", n_datasets, n_ms_files)
        logger.info("Saving index to %r", cache_name)
        try:
            with open(cache_name, 'wt') as fh:
                json.dump(index, fh, indent=2, sort_keys=True)
        except IOError:
            logger.error("Failed to write index to disk", exc_info=True)
        return index

    def _traverse_files(self, file_types, index):
        raise NotImplementedError()

    def find_ms_files_for(self, dataset, run_name):
        result = []
        members = self.index.get(dataset)
        if run_name is None:
            return members

        if members is None:
            raise DatasetNotAvailable((dataset, run_name))

        for member in members:
            base = member.split("/")[-1]
            if base.startswith(run_name):
                result.append(member)
            elif base.endswith("json"):
                if base.startswith(run_name.rsplit(".", 1)[0]):
                    result.append(member)
        # Validate that there is only one MS file for the name
        i = 0
        for res in sorted(result):
            if re.search(r"-byte-offsets.json$", res):
                continue
            if re.search(r"-idx.json$", res):
                continue
            i += 1
        if i > 1:
            raise ValueError("Multiple MS Data Files Found")
        return self._find_mz_files(result, first=True), self._find_index_files(result, first=True)

    def _find_mz_files(self, namelist, first=False):
        pattern = re.compile(r"(?:%s)(?:.gz)?$" %
                             "|".join(self.valid_file_types))
        names = []
        for name in namelist:
            if pattern.search(name):
                if first:
                    return name
                names.append(name)
        if first and not names:
            return None
        return names

    def _find_index_files(self, namelist, first=False):
        names = []
        for name in namelist:
            if re.search(r"-byte-offsets.json$", name):
                if first:
                    return name
                names.append(name)
        if first and not names:
            return None
        return names

    def _opener(self, uri, block_size=None, **kwargs):
        raise NotImplementedError()

    def open_ms_file(self, ms_file_uri, index_uri=None):
        if ms_file_uri in self.cache:
            logger.info("Found %r in the cache", ms_file_uri)
            return self.cache[ms_file_uri]
        logger.info("Opening %r", ms_file_uri)
        base_name =  ms_file_uri.split("/")[-1]
        if base_name.split('.')[-1] in self.special_openers:
            reader = ms_deisotope.MSFileLoader(ms_file_uri)
        else:
            mzml_fh = self._opener(ms_file_uri)
            if ms_file_uri.endswith("gz"):
                mzml_fh = idzip.IdzipFile(fileobj=mzml_fh)
            index_fh = None
            if index_uri is not None:
                index_fh = self._opener(index_uri, block_size=BLOCK_SIZE)
            else:
                logger.warning("Byte offset index is missing for %r", ms_file_uri)
            reader = ms_deisotope.MSFileLoader(mzml_fh, index_file=index_fh)
        logger.info("Finished opening %r", ms_file_uri)
        lock = RLock()
        self.cache[ms_file_uri] = (reader, lock)
        return reader, lock

    def list_collections(self):
        return list(self.index.keys())

    def list_datasets(self):
        return {
            k: [f.split("/")[-1] for f in self._find_mz_files(v)] for k, v in self.index.items()
        }

    def get_scan_from_usi(self, usi, reader):
        if usi.scan_identifier_type == 'scan':
            index = int(usi.scan_identifier) - 1
        elif usi.scan_identifier_type == "index":
            index = int(usi.scan_identifier)
        else:
            raise UnrecognizedIndexFlag(usi.scan_identifier_type)
        scan = reader.get_scan_by_index(index)
        return scan

    def process_scan(self, scan, reader):
        # No processing
        return scan

    def enumerate_spectra_for(self, dataset, ms_run=None):
        raise NotImplementedError()

    def handle_usi(self, usi, convert_json=True):
        """Receive a USI, parse it into its constituent pieces, locate
        the relevant spectrum, and return it formatted message.

        Parameters
        ----------
        usi : str
            The identifier of the spectrum to be retrieved
        convert_json : bool, optional
            Whether to format the spectrum as JSON or not. Defaults to :const:`True`.

        Returns
        -------
        spectrum : dict or Scan
            The spectrum, rendered according to the `convert_json` or not.
        """
        usi = USI.parse(str(usi))
        mzml_uri, index_uri = self.find_ms_files_for(usi.dataset, usi.datafile)
        reader, lock = self.open_ms_file(mzml_uri, index_uri)
        with lock:
            scan = self.get_scan_from_usi(usi, reader)
            scan = self.process_scan(scan, reader)
            if convert_json:
                payload = self.build_response(usi, scan)
                return payload
            else:
                return scan

    def __call__(self, usi):
        return self.handle_usi(usi)


class FSMassSpectraDataArchive(MassSpectraDataArchiveBase):
    def __init__(self, base_uri, cache_size=None, **kwargs):
        self.valid_file_types = self.valid_file_types[:]
        from ms_deisotope.data_source.thermo_raw_net import determine_if_available
        if determine_if_available():
            self.valid_file_types.append("raw")
            self.special_openers = set(self.special_openers) | {"raw"}
        super(FSMassSpectraDataArchive, self).__init__(
            base_uri, cache_size=cache_size, **kwargs)

    def _opener(self, uri, block_size=None, **kwargs):
        return open(uri, 'rb')

    def _datafile_uri(self, dataset_id, data_file):
        return os.path.normpath('/'.join([self.base_uri, data_file])).replace(os.sep, '/')

    def _traverse_files(self, file_types, index):
        for prefix, dirs, files in os.walk(self.base_uri):
            if os.sep in prefix:
                tokens = prefix.split(os.sep)
            else:
                tokens = prefix.split("/")
            if tokens[-1] == '':
                base = tokens[-2]
            else:
                base = tokens[-1]

            for f in files:
                try:
                    ext = f.rsplit(".", 1)[1]
                except IndexError:
                    continue
                if ext == 'gz':
                    ext = f.rsplit(".", 2)[1]
                if ext in file_types:
                    index[base].append(
                        os.path.normpath(
                            os.path.join(prefix, f)).replace(
                                os.sep, "/"))


class S3MassSpectraDataArchive(MassSpectraDataArchiveBase):
    def __init__(self, base_uri, cache_size=None, s3_args=None, **kwargs):
        if s3_args is None:
            s3_args = {}
        self._tls = local()
        self.s3_args = s3_args
        super(S3MassSpectraDataArchive, self).__init__(
            base_uri=base_uri, cache_size=cache_size, **kwargs)

    @property
    def s3_service(self):
        try:
            return self._tls.s3_service
        except (AttributeError, KeyError):
            from s3fs import S3FileSystem
            self._tls.s3_service = S3FileSystem(**self.s3_args)
            return self._tls.s3_service

    def _opener(self, uri, block_size=None, **kwargs):
        if block_size is None:
            block_size = BLOCK_SIZE
        return self.s3_service.open(uri, block_size=block_size)

    def _datafile_uri(self, dataset_id, data_file):
        return "s3://" + data_file

    def _traverse_files(self, file_types, index):
        for prefix, _dirs, files in self.s3_service.walk(self.base_uri):
            for f in files:
                query = f
                if query.endswith(".gz"):
                    query = query[:-3]
                ext = query.rsplit(".", 1)[-1]
                for t in file_types:
                    if t == ext:
                        index[prefix.replace(self.base_uri[5:], '')].append(
                            prefix + '/' + f)
                        break
        for group, members in list(index.items()):
            if "/" in group:
                base, _rest = group.split("/", 1)
                index[base].extend(members)
                index.pop(group)


class PeakPickingScanProcessingMixin(object):
    def __init__(self, *args, **kwargs):
        super(PeakPickingScanProcessingMixin, self).__init__(*args, **kwargs)

    def process_scan(self, scan, reader):
        scan = super(PeakPickingScanProcessingMixin,
                     self).process_scan(scan, reader)
        scan.pick_peaks()
        return scan


class DeconvolutingScanProcessingMixin(PeakPickingScanProcessingMixin):
    def __init__(self, *args, **kwargs):
        self.deconvolution_args = kwargs.get('deconvolution_args', {})
        self.deconvolution_args.setdefault("averagine", ms_deisotope.peptide)
        self.deconvolution_args.setdefault(
            "scorer", ms_deisotope.MSDeconVFitter(0))
        self.deconvolution_args.setdefault("truncate_after", 0.8)
        super(DeconvolutingScanProcessingMixin, self).__init__(*args, **kwargs)

    def process_scan(self, scan, reader):
        scan = super(DeconvolutingScanProcessingMixin, self).process_scan(scan, reader)
        scan.deconvolute(**self.deconvolution_args)
        return scan



import random
from typing import List, Sequence, Set
import numpy as np

from collections import OrderedDict

from ms_peak_picker import PeakLike as PeakBase

from ms_deisotope.utils import uid


class FeatureBase(object):
    nodes: 'LCMSFeatureTreeList'

    def __init__(self, nodes):
        self.nodes = LCMSFeatureTreeList(nodes)

    def find_time(self, time_point):
        return self.nodes.find_time(time_point)


class LCMSFeature(FeatureBase):
    created_at = "new"

    _total_intensity: float
    _mz: float
    _last_mz: float
    _times: np.ndarray
    _peaks: Sequence[List[PeakBase]]
    _start_time: float
    _end_time: float
    adducts: List
    used_as_adduct: List
    feature_id: int
    _peak_averager: 'RunningWeightedAverage'

    def __init__(self, nodes=None, adducts=None, used_as_adduct=None, feature_id=None):
        if nodes is None:
            nodes = []
        if adducts is None:
            adducts = []
        if used_as_adduct is None:
            used_as_adduct = []
        if feature_id is None:
            feature_id = uid()

        FeatureBase.__init__(self, nodes)
        self._total_intensity = None
        self._mz = None
        self._last_mz = 0.0
        self._times = None
        self._peaks = None
        self._start_time = None
        self._end_time = None
        self.adducts = adducts
        self.used_as_adduct = used_as_adduct
        self.feature_id = feature_id
        self._initialize_averager()

    def _reaverage_and_update(self):
        self._peak_averager.reset()
        self._peak_averager.feed_from_feature(self)
        self._update_from_averager()

    def _initialize_averager(self):
        self._peak_averager = RunningWeightedAverage()

    def _update_from_averager(self):
        self._mz = self._last_mz = self._peak_averager.current_mean

    def invalidate(self, reaverage=False):
        self._invalidate(reaverage)

    def _invalidate(self, reaverage=False):
        self._total_intensity = None
        self._last_mz = self._mz if self._mz is not None else 0.
        self._mz = None
        self._times = None
        self._peaks = None
        self._start_time = None
        self._end_time = None

        if reaverage:
            self._peak_averager.reset()
            if len(self) > 0:
                self._feed_peak_averager()

    @property
    def total_signal(self):
        if self._total_intensity is None:
            total = 0.
            for node in self.nodes:
                total += node.total_intensity()
            self._total_intensity = total
        return self._total_intensity

    @property
    def intensity(self):
        return self.total_signal

    @property
    def neutral_mass(self):
        return self.mz

    @property
    def mz(self):
        if self._mz is None:
            self._reaverage_and_update()
        return self._mz

    @property
    def times(self):
        if self._times is None:
            self._times = np.array(
                [node.time for node in self.nodes])
        return self._times

    @property
    def peaks(self):
        if self._peaks is None:
            self._peaks = tuple(node.peaks for node in self.nodes)
        return self._peaks

    @property
    def start_time(self):
        if self._start_time is None:
            self._start_time = self.nodes[0].time
        return self._start_time

    @property
    def end_time(self):
        if self._end_time is None:
            self._end_time = self.nodes[-1].time
        return self._end_time

    def overlaps_in_time(self, interval) -> bool:
        cond = ((self.start_time <= interval.start_time and self.end_time >= interval.end_time) or (
            self.start_time >= interval.start_time and self.end_time <= interval.end_time) or (
            self.start_time >= interval.start_time and self.end_time >= interval.end_time and
            self.start_time <= interval.end_time) or (
            self.start_time <= interval.start_time and self.end_time >= interval.start_time) or (
            self.start_time <= interval.end_time and self.end_time >= interval.end_time))
        return cond

    def spans_in_time(self, time: float) -> bool:
        return self.start_time <= time <= self.end_time

    def as_arrays(self):
        rts = np.array(
            [node.time for node in self.nodes], dtype=np.float64)
        signal = np.array([node.total_intensity()
                           for node in self.nodes], dtype=np.float64)
        return rts, signal

    def __len__(self):
        return len(self.nodes)

    def __repr__(self):
        return "%s(%0.4f, %0.2f, %0.2f)" % (
            self.__class__.__name__, self.mz, self.start_time, self.end_time)

    def _copy_chunk(self, nodes, *args, **kwargs):
        x = self.__class__(LCMSFeatureTreeList(nodes))
        x.used_as_adduct = list(self.used_as_adduct)
        return x

    def split_at(self, time: float):
        _, i = self.nodes.find_time(time)
        if self.nodes[i].time < time:
            i += 1
        return LCMSFeature(self.nodes[:i]), LCMSFeature(self.nodes[i:])

    def split_sparse(self, delta_rt: float=1.) -> List['LCMSFeature']:
        chunks = []
        current_chunk = []
        last_rt = self.nodes[0].time

        for node in self.nodes:
            if (node.time - last_rt) > delta_rt:
                x = self._copy_chunk(current_chunk)

                chunks.append(x)
                current_chunk = []

            last_rt = node.time
            current_chunk.append(node)

        x = self._copy_chunk(current_chunk)

        chunks.append(x)
        for chunk in chunks:
            chunk.created_at = self.created_at

        for member in chunks:
            for other in chunks:
                if member == other:
                    continue
                assert not member.overlaps_in_time(other)

        return chunks

    def truncate_before(self, time: float):
        _, i = self.nodes.find_time(time)
        if self.nodes[i].time < time:
            i += 1
        self.nodes = LCMSFeatureTreeList(self.nodes[i:])
        self._invalidate()

    def truncate_after(self, time: float):
        _, i = self.nodes.find_time(time)
        if self.nodes[i].time < time:
            i += 1
        self.nodes = LCMSFeatureTreeList(self.nodes[:i])
        self._invalidate()

    def clone(self, deep=False, cls=None):
        if cls is None:
            cls = self.__class__
        c = cls(self.nodes.clone(deep=deep), list(
            self.adducts), list(self.used_as_adduct))
        c.feature_id = self.feature_id
        c.created_at = self.created_at
        return c

    def insert_node(self, node):
        self._peak_averager.update(node.members)
        self.nodes.insert_node(node)
        self._invalidate()

    def insert(self, peak, time):
        self._peak_averager.add(peak)
        self.nodes.insert(time, [peak])
        self._invalidate()

    def merge(self, other):
        new = self.clone()
        for node in other.nodes:
            node = node.clone()
            new.insert_node(node)
        new.created_at = "merge"
        return new

    @property
    def apex_time(self):
        rt, intensity = self.as_arrays()
        return rt[np.argmax(intensity)]

    def __iter__(self):
        return iter(self.nodes)

    def __eq__(self, other):
        if other is None:
            return False
        if len(self) != len(other):
            return False
        else:
            if abs(self.start_time - other.start_time) > 1e-4:
                return False
            elif abs(self.end_time - other.end_time) > 1e-4:
                return False
            else:
                for a, b in zip(self, other):
                    if a != b:
                        return False
                return True

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.mz, self.start_time, self.end_time))

    def __getitem__(self, i):
        return self.nodes[i]


class LCMSFeatureTreeList(object):
    roots: List['LCMSFeatureTreeNode']
    _node_id_hash: Set[int]

    def __init__(self, roots=None):
        if roots is None:
            roots = []
        self.roots = list(roots)
        self._node_id_hash = None

    def _invalidate(self):
        self._node_id_hash = None

    def find_time(self, time):
        if len(self.roots) == 0:
            raise ValueError()
        lo = 0
        hi = len(self.roots)
        while lo != hi:
            i = (lo + hi) // 2
            node = self.roots[i]
            if node.time == time:
                return node, i
            elif (hi - lo) == 1:
                return None, i
            elif node.time < time:
                lo = i
            elif node.time > time:
                hi = i

    def _build_node_id_hash(self):
        self._node_id_hash = set()
        for node in self.unspool():
            self._node_id_hash.add(node.node_id)

    @property
    def node_id_hash(self):
        if self._node_id_hash is None:
            self._build_node_id_hash()
        return self._node_id_hash

    def insert_node(self, node: 'LCMSFeatureTreeNode'):
        self._invalidate()
        try:
            root, i = self.find_time(node.time)
            if root is None:
                if i != 0:
                    self.roots.insert(i + 1, node)
                else:
                    slot = self.roots[i]
                    if slot.time < node.time:
                        i += 1
                    self.roots.insert(i, node)
            else:
                root.add(node)
            return i
        except ValueError:
            self.roots.append(node)
            return 0

    def insert(self, time, peaks):
        node = LCMSFeatureTreeNode(time, peaks)
        return self.insert_node(node)

    def extend(self, iterable):
        for peaks, time in iterable:
            self.insert(time, peaks)

    def __getitem__(self, i):
        return self.roots[i]

    def __len__(self):
        return len(self.roots)

    def __iter__(self):
        return iter(self.roots)

    def clone(self, deep=False):
        return LCMSFeatureTreeList(node.clone(deep=deep) for node in self)

    def unspool(self):
        out_queue = []
        for root in self:
            stack = [root]
            while len(stack) != 0:
                node = stack.pop()
                out_queue.append(node)
                stack.extend(node.children)
        return out_queue

    def common_nodes(self, other):
        return len(self.node_id_hash & other.node_id_hash)

    def __repr__(self):
        return "LCMSFeatureTreeList(%d nodes, %0.2f-%0.2f)" % (
            len(self), self[0].time, self[-1].time)


class LCMSFeatureTreeNode(object):

    __slots__ = ["time", "members",
                 "_most_abundant_member", "_mz", "node_id"]

    time: float
    members: List[PeakBase]
    _most_abundant_member: PeakBase
    _mz: float
    node_id: int

    def __init__(self, time=None, members=None):
        if members is None:
            members = []
        self.time = time
        self.members = members
        self._most_abundant_member = None
        self._mz = 0
        self._recalculate()
        self.node_id = uid()

    def clone(self, deep=False):
        if deep:
            peaks = [peak.clone() for peak in self.members]
        else:
            peaks = list(self.members)
        node = self.__class__(
            self.time, peaks)
        node.node_id = self.node_id
        return node

    def _unspool_strip_children(self):
        node = self.__class__(
            self.time, list(self.members))
        yield node

    def _calculate_most_abundant_member(self):
        if len(self.members) == 1:
            self._most_abundant_member = self.members[0]
        else:
            if len(self.members) == 0:
                self._most_abundant_member = None
            else:
                self._most_abundant_member = max(
                    self.members, key=lambda x: x.intensity)

    def _recalculate(self):
        self._calculate_most_abundant_member()
        self._mz = self._most_abundant_member.mz

    def __getstate__(self):
        return (self.time, self.members, self.node_id)

    def __setstate__(self, state):
        self.time, self.members, self.node_id = state
        self._recalculate()

    def __reduce__(self):
        return self.__class__, (self.time, []), self.__getstate__()

    @property
    def mz(self):
        if self._mz is None:
            if self._most_abundant_member is not None:
                self._mz = self._most_abundant_member.mz
        return self._mz

    def add(self, node: 'LCMSFeatureTreeNode', recalculate=True):
        if self.node_id == node.node_id:
            raise ValueError("Duplicate Node %s" % node)
        self.members.extend(node.members)
        if recalculate:
            self._recalculate()

    def _total_intensity_members(self):
        total = 0.
        for peak in self.members:
            total += peak.intensity
        return total

    def max_intensity(self):
        return self._most_abundant_member.intensity

    def total_intensity(self):
        return self._total_intensity_members()

    def __eq__(self, other):
        return self.members == other.members and abs(
            self.time - other.time) < 1e-4

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.uid)

    @property
    def peaks(self):
        peaks = list(self.members)
        return peaks

    def __repr__(self):
        return "%s(%f, %0.4f %s|%d)" % (
            self.__class__.__name__,
            self.time, self.peaks[0].mz, "",
            len(self.members))


class EmptyFeature(FeatureBase):

    def __init__(self, mz):
        FeatureBase.__init__(self, [])
        self.mz = mz
        self.start_time = 0
        self.end_time = float('inf')

    def __len__(self):
        return 0

    @property
    def times(self):
        return np.array([])

    def __iter__(self):
        return iter([])

    def __repr__(self):
        return "EmptyFeature(%f)" % (self.mz,)


class FeatureSetIterator(object):

    def __init__(self, features, start_time=None, end_time=None):
        self.features = features
        self.real_features = [
            f for f in features if f is not None and not isinstance(f, EmptyFeature)]
        self.start_time = max(
            [f.start_time for f in self.real_features]
        ) if start_time is None else start_time
        self.end_time = min(
            [f.end_time for f in self.real_features]
        ) if end_time is None else end_time

        self.index_list = [0 for f in self.features]
        self.init_indices()
        self.last_seen_time = -1

    def init_indices(self):
        for i, f in enumerate(self.features):
            if f is None:
                self.index_list[i] = 0
                continue
            try:
                node, ix = f.find_time(self.start_time)
            except ValueError:
                if not isinstance(f, EmptyFeature):
                    raise
                else:
                    ix = 0
            self.index_list[i] = ix

    def get_next_time(self):
        time = float('inf')
        for i, ix in enumerate(self.index_list):
            try:
                f = self.features[i]
                if f is None:
                    continue
                if ix < len(f):
                    ix_time = self.features[i][ix].time
                    if ix_time < time and ix_time < self.end_time:
                        time = ix_time
            except IndexError:
                continue
        if time == float('inf'):
            return None
        else:
            return time

    def has_more(self):
        j = 0
        for i, f in enumerate(self.features):
            if f is None:
                j += 1
                continue
            ix = self.index_list[i]
            done = ix >= len(f)
            if not done:
                at_end_time = f[ix].time >= self.end_time
            else:
                at_end_time = done
            if done or at_end_time:
                j += 1
        return j != len(self.features)

    def get_peaks_for_time(self, time):
        peaks = []
        i = 0
        for feature in self.features:
            if feature is not None:
                try:
                    node, ix = feature.nodes.find_time(time)
                    if ix == self.index_list[i] and self.index_list[i] == 0 and node is None:
                        pass
                    elif ix >= self.index_list[i]:
                        self.index_list[i] += 1
                except ValueError:
                    if not isinstance(feature, EmptyFeature):
                        raise
                    node = None
                if node is not None:
                    peaks.append(
                        node.members[0])
                else:
                    peaks.append(None)
            else:
                peaks.append(None)
            i += 1
        return peaks

    @property
    def current_time(self):
        return self.last_seen_time

    def get_next_value(self):
        time = self.get_next_time()
        if time is None:
            raise StopIteration()
        self.last_seen_time = time
        peaks = self.get_peaks_for_time(time)
        return peaks

    def __next__(self):
        return self.get_next_value()

    def next(self):
        return self.get_next_value()

    def __iter__(self):
        return self

    def __repr__(self):
        return "FeatureSetIterator(%d/%d, %0.3f-%0.3f)" % (
            len(self.real_features), len(self.features), self.start_time, self.end_time)


try:
    has_c = True
    _LCMSFeatureTreeList = LCMSFeatureTreeList
    _LCMSFeatureTreeNode = LCMSFeatureTreeNode
    _LCMSFeature = LCMSFeature
    _FeatureSetIterator = FeatureSetIterator
    _EmptyFeature = EmptyFeature
    _FeatureBase = FeatureBase
    from ms_deisotope._c.feature_map.lcms_feature import (
        LCMSFeatureTreeList,
        LCMSFeatureTreeNode,
        LCMSFeature,
        FeatureSetIterator,
        EmptyFeature,
        FeatureBase
    )
except ImportError:
    has_c = False


class NodeFeatureSetIterator(FeatureSetIterator):

    def get_peaks_for_time(self, time):
        peaks = []
        for feature in self.features:
            if feature is not None:
                try:
                    node, ix = feature.nodes.find_time(time)
                except ValueError:
                    if not isinstance(feature, EmptyFeature):
                        raise
                    node = None
                if node is not None:
                    peaks.append(node)
                else:
                    peaks.append(None)
            else:
                peaks.append(None)
        return peaks


def iterpeaks(feature):
    for node in feature:
        for peak in node.members:
            yield peak


class RunningWeightedAverage(object):

    def __init__(self, iterable=None):
        self.total_weight = 0
        self.current_mean = 0
        self.current_count = 0

        if iterable is not None:
            self.update(iterable)

    def add(self, peak):
        if peak.intensity == 0:
            if self.current_mean == 0 and self.total_weight == 0:
                self.current_mean = peak.mz
                self.total_weight = 1
            else:
                return
        agg = (self.total_weight * self.current_mean) + \
            (peak.mz * peak.intensity)
        self.total_weight += peak.intensity
        self.current_mean = agg / self.total_weight
        self.current_count += 1
        return self

    def feed_from_feature(self, feature: LCMSFeature):
        for node in feature:
            self.update(node.members)
        return len(feature)

    def update(self, iterable):
        for x in iterable:
            self.add(x)
        return self

    def __repr__(self):
        return "RunningWeightedAverage(%r, %d)" % (self.current_mean, self.current_count)


class SimpleLCMSFeature(object):
    def __init__(self, storage=None):
        if storage is None:
            storage = OrderedDict()
        self.storage = storage

    def as_arrays(self):
        return (
            np.array(self.storage.keys()),
            np.array(self.storage.values()))

    @property
    def total_signal(self):
        return sum(self.storage.values())

    @property
    def apex_time(self):
        rt, intensity = self.as_arrays()
        return rt[np.argmax(intensity)]

    @property
    def start_time(self):
        return tuple(self.storage.keys())[0]

    @property
    def end_time(self):
        return tuple(self.storage.keys())[-1]


def smooth_feature(feature, level=1):
    from ms_deisotope.feature_map.profile_transform import smooth_leveled
    time, intensity = feature.as_arrays()
    smoothed = smooth_leveled(time, intensity, level)
    for i, node in enumerate(feature):
        peaks = node.members
        if len(peaks) > 1:
            peaks.sort(key=lambda x: x.intensity, reverse=True)
            peak = peaks[0]
            for other in peaks[1:]:
                peak.intensity += other.intensity
            node.members = [peak]
        else:
            peak = node.members[0]
        peak.intensity = smoothed[i]
    feature.invalidate(True)
    return feature

try:
    from ms_deisotope._c.feature_map.lcms_feature import RunningWeightedAverage
except ImportError:
    has_c = False



"""Represent controlled vocabulary (CV) terms and provides the machinery used
for building compile-time collections of CV terms from the PSI-MS controlled
vocabulary.
"""
import sys
import textwrap

from six import string_types as basestring


class Term(object):
    """Represents a single controlled vocabulary term.

    :class:`Term` objects are comparable and hashable.

    Attributes
    ----------
    name: str
        The human-readable name for the term
    id: str
        A namespace-qualified unique alphanumeric identifier for the term
    description: str
        A longer description of the concept the term represents.
    category: str
        The name of the broad class of terms this term belongs to
    specialization: list
        A sequence of increasingly specific categories that this term
        belongs to
    """
    __slots__ = ("name", "id", "description", "category", "specialization")

    def __init__(self, name, id, description, category, specialization):
        self.name = name
        self.id = id
        self.description = description
        self.category = category
        self.specialization = specialization

    def __iter__(self):
        yield self.name
        yield self.id
        yield self.description
        yield self.category
        yield self.specialization

    def _asdict(self):
        return {
            "name": self.name,
            "id": self.id,
            "description": self.description,
            "category": self.category,
            "specialization": self.specialization
        }

    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, basestring):
            return self.name == other or self.id == other
        else:
            if (self.name != other.name) or (self.id != other.id) or\
                    (self.description != other.description) or\
                    (self.category != other.category) or\
                    (self.specialization != other.specialization):
                result = False
            else:
                result = True
            return result

    def __str__(self):
        return str(self.name)

    def __repr__(self):
        text = "(%s)" % ', '.join("%s=%r" % (k, v) for k, v in self._asdict().items() if k != 'description')
        return self.__class__.__name__ + text

    def __reduce__(self):
        return self.__class__, (None, None, None, None, None), self.__getstate__()

    def __getstate__(self):
        return tuple(self)

    def __setstate__(self, d):
        if len(d) == 4:
            self.name, self.id, self.category, self.specialization = d
        else:
            self.name, self.id, self.description, self.category, self.specialization = d

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.name)

    def is_a(self, term):
        """Test whether this entity is exactly **term** or a specialization
        of **term**

        Parameters
        ----------
        term : str or :class:`~.Term`
            The entity to compare to

        Returns
        -------
        bool
        """
        return getattr(term, 'name', term) == self.name or term in self.specialization


class TermSet(object):
    """A collection that mocks a list and a dictionary for controlled vocabulary terms

    Attributes
    ----------
    by_id : dict
        Mapping from :attr:`Term.id` to :class:`Term`
    by_name : dict
        Mapping from :attr:`Term.name` to :class:`Term`
    terms : list
        List of :class:`Term` objects
    """

    def __init__(self, terms):
        self.terms = list(terms)
        self.by_name = {
            t.name: t for t in self.terms
        }
        self.by_id = {
            t.id: t for t in self.terms
        }

    def __iter__(self):
        return iter(self.terms)

    def __len__(self):
        return len(self.terms)

    def __add__(self, other):
        return self.__class__(list(self) + list(other))

    def __contains__(self, term):
        return term in self.terms or term in self.by_id or term in self.by_name

    def keys(self):
        """Returns the list of keys that this collection recognizes, over
        all of its wrapped mappings.
        """
        return set(self.by_id.keys()) | set(self.by_name.keys())

    def get(self, key, default=None):
        """Return the value of ``key`` or ``default`` if it is not
        found

        Parameters
        ----------
        key: str
            The key to look for
        default: object
            The value to return if ``key`` is not found. Defaults to :const:`None`.

        Returns
        -------
        object
        """
        try:
            return self[key]
        except (KeyError, IndexError):
            return default

    def __getitem__(self, k):
        if isinstance(k, int):
            return self.terms[k]
        try:
            return self.by_id[k]
        except KeyError:
            pass
        try:
            return self.by_name[k]
        except KeyError:
            pass
        raise KeyError(k)


def _unique_list(items):  # pragma: no cover
    seen = set()
    out = []
    for x in items:
        if x in seen:
            continue
        seen.add(x)
        out.append(x)
    return out


class MappingProxy(object):
    """An object that proxies :meth:`__getitem__` to another object
    which is loaded lazily through a callable :attr:`loader`
    """
    def __init__(self, loader):
        assert callable(loader)
        self.loader = loader
        self.mapping = None

    @property
    def metadata(self):
        self._ensure_mapping()
        return self.mapping.metadata

    def _ensure_mapping(self):
        if self.mapping is None:
            self.mapping = self.loader()

    def __getitem__(self, key):
        self._ensure_mapping()
        return self.mapping[key]


def _lazy_load_psims():
    try:
        from psims.controlled_vocabulary.controlled_vocabulary import load_psims
        cv = load_psims()
    except Exception:  # pragma: no cover
        cv = None
    return cv


cv_psims = MappingProxy(_lazy_load_psims)


def _clean_definition(text):
    if text.startswith('"'):
        text = text.rsplit(" ", 1)[0]
        text = text[1:-1]
    return text


def type_path(term, seed):  # pragma: no cover
    """Traverse is-a relationships from more specialized to
    less specialized until the root term type has been found,
    accumulating types along the way.

    Parameters
    ----------
    term: :class:`psims.controlled_vocabulary.Term`
        The term to traverse
    seed: object
        Unused

    Returns
    -------
    list
    """
    path = []
    i = 0
    steps = []
    if not isinstance(term.is_a, (list, tuple)):
        steps.append(term.is_a)
    else:
        steps.extend(t for t in term.is_a)
    while i < len(steps):
        step = steps[i]
        i += 1
        try:
            path.append(step.comment)
            term = cv_psims[step.accession]
        except AttributeError as err:
            print(step, err)
            path.append(step)
            term = cv_psims[step]
        try:
            if not isinstance(term.is_a, (list, tuple)):
                steps.append(term.is_a)
            else:
                steps.extend(t for t in term.is_a)
        except KeyError:
            continue
    return _unique_list(path)


def render_list(seed, list_name=None, term_cls_name="Term", writer=None):  # pragma: no cover
    """A code generator for rendering static lists of :class:`Term`-like objects
    from PSI-MS.
    """
    if writer is None:
        writer = sys.stdout.write
    component_type_list = [seed]
    i = 0
    seen = set()
    if list_name is None:
        list_name = seed.replace(" ", "_") + 's'
    template = (
        "    %s(%r, %r,\n    %s,\n"
        "       %r,\n       %r), \n")

    def _wraplines(text, width=60, indent='        '):
        lines = textwrap.wrap(text, width=60)
        n_lines = len(lines)
        for i in range(n_lines - 1):
            lines[i] = lines[i] + ' '
        lines = map(repr, lines)
        return indent[:-1] + '(' + ('\n' + indent).join(lines) + ')'
    writer("# CV Version: %s\n" % cv_psims.metadata['data-version'])
    writer("%s = TermSet([\n" % (list_name,))
    while i < len(component_type_list):
        component_type = component_type_list[i]
        i += 1
        for term in sorted(cv_psims[component_type].children, key=lambda x: x.id):
            if term.name in seen:
                continue
            seen.add(term.name)
            writer(template % (
                term_cls_name, term.name, term.id, _wraplines(_clean_definition(term.get("def", ''))),
                component_type_list[0], type_path(term, seed)))
            if term.children:
                component_type_list.append(term.name)
    writer("])\n")


__all__ = [
    "Term", "cv_psims", "render_list",
    "MappingProxy"
]
